; ATtiny13 based dice
; Copyright (c) 2017 F. Stefanec
;
; Fuse settings: HFUSE=0xFF, LFUSE=0x7F
; -> 128kHz internal oscillator
;
; A button is connected between PB3 and GND.
;
;   [PB3]---[button]---[ground]
;
; The dice display consists of 7 LEDs. 
;
; There is a resistor (270-390 ohm for 3V power, 820-1200ohm for 5V power) in series with each LED.
;   [microcontroller pin]---[resistor]---[LED anode | LED cathode]---[ground]
;
; The LEDs are positioned like this:
;
;     PB2             PB1
;
; 
;
;     PB0     PB4     PB0
;
;
;
;     PB1             PB2


.include "tn13def.inc"

.def ZERO = R0 ; zero register
.def REG = R16 ; general purpose register
.def IREG = R17 ; general purpose register (for use inside ISR)
.def STAT = R18 ; status register
.def COUNTER = R19 ; counter
.def REG2 = R20 ; another general purpose register
.def CR_NORMAL = R21 ; MCUCR normal value register
.equ STAT_SLEEP = 3


.equ MCUCR_NORMAL = 0 ; MCU control register normal value
.equ MCUCR_SLEEPE = (1<<SM1)|(1<<SE) ; MCUCR with Sleep Enable

.equ BTN_PIN = PB3
.equ BTN_PCINT = PCINT3

; Port B
; --- --- PB5(reset) PB4 PB3 (button) PB2 PB1 PB0

.equ ALLOFF = 	0b00101000
.equ ALLON = 	0b00111111
.equ DICE1 = 	0b00111000
.equ DICE2 = 	0b00101010
.equ DICE3 = 	0b00111100
.equ DICE4 = 	0b00101110
.equ DICE5 = 	0b00111110
.equ DICE6 = 	0b00101111

.CSEG

.ORG 0 ; jump to main loop
RJMP MAIN

.ORG PCI0addr ; pin change interrupt vector
RJMP BUTTON

; Button state change interrupt
BUTTON:	OUT 	MCUCR, CR_NORMAL
	CLI
	OUT	PCMSK, ZERO
	SBR	STAT, STAT_SLEEP 
	SBIC	PINB, BTN_PIN ; return if button value = high
	RETI
	LDI	IREG, MCUCR_NORMAL
	OUT	MCUCR, IREG
CLOOP:	LDI 	IREG, ALLON ; Light up all LEDs
	OUT	PORTB, IREG
	; Counter loop:
LOOP1:	NOP 	; 2 cycle delay - so the loop takes the same time with all possible values of COUNTER reigster
	NOP
LOOP2:	SBIC	PINB, BTN_PIN ; jump out of loop if button not pressed
	RJMP 	LOOP3
	INC	COUNTER ; increment counter
	CPI 	COUNTER, 6	
	BRLO	LOOP1
	LDI 	COUNTER, 0
	RJMP 	LOOP2
LOOP3:	RCALL 	DELAY1 ; wait 20ms (button bounce protection)
	RCALL 	DELAY1
	; Output LED values
	CPI	COUNTER, 0 ; compare 0-5 with counter value, skip if not equal, else load dice LED values into IREG
	BRNE	C0e
	LDI	IREG, DICE1
C0e: 	CPI 	COUNTER, 1
	BRNE	C1e
	LDI	IREG, DICE2
C1e: 	CPI 	COUNTER, 2
	BRNE	C2e
	LDI	IREG, DICE3
C2e: 	CPI 	COUNTER, 3
	BRNE	C3e
	LDI	IREG, DICE4
C3e: 	CPI 	COUNTER, 4
	BRNE	C4e
	LDI	IREG, DICE5
C4e: 	CPI 	COUNTER, 5
	BRNE	C5e
	LDI	IREG, DICE6
C5e:	OUT	PORTB, IREG ; output dice LED values (write into PORTB)
	; Button push handler & wait loop (20 sec.)
	; REG2 is used as time counter (for button press length, REG & COUNTER are used for wait loop timing)
	PUSH 	REG2 ; push registers on stack
	CLR	REG2
	PUSH 	REG
	PUSH	COUNTER
	LDI	COUNTER, 8 ; two wait loops (250*8*10 = 20000 ms, 20 sec delay, 10 ms steps)
WLO1:	LDI 	REG, 250
WLO2:	RCALL	DELAY1
	SBIC	PINB, BTN_PIN
	RJMP	BTN_NP ; jump to BTN_NP if button not pressed
	INC 	REG2 ; increase press time counter
	SBRC	REG2, 7 ; don't let REG2 get over 128
	LDI	REG2, 127
	CPI	REG2, 40 ; long press? (>400 ms)
	BRLO	NOLP ; jump to NOLP if no long press
	POP	COUNTER ; if long press occurred, pop pushed registers & repeat loop
	POP 	REG
	POP 	REG2
	RJMP 	CLOOP
NOLP: 	RJMP 	WLO3 ; end wait loop
BTN_NP: CPI	REG2, 2 ; short press? (>20, <400ms)
	BRLO	NOSP ; jump to NOSP if no short press
	RJMP	PEND
NOSP:	LDI	REG2, 0
WLO3:	DEC	REG ; end of wait loop program, decrease loop counter registers and branch if not zero
	BRNE	WLO2 
	DEC	COUNTER
	BRNE 	WLO1
PEND:	POP 	COUNTER ; end, pop registers, clear LEDs... and return
	POP 	REG
	POP 	REG2
	LDI	IREG, ALLOFF ; turn off all LEDs
	OUT	PORTB, IREG
	RCALL	DELAY1
	RCALL 	DELAY1
WFBR:	SBIS	PINB, BTN_PIN ; wait for button release
	RJMP 	WFBR
	NOP
	OUT	GIFR, ZERO ; clear general interrupt flag register
	SBR	STAT, STAT_SLEEP
	RETI
	

; Main program
MAIN:	CLI
	CLR	ZERO ; clear zero register
	LDI 	CR_NORMAL, MCUCR_NORMAL ; MCU control register normal value register
	LDI	REG, low(RAMEND) ; set stack pointer to end of RAM, set I/O registers
	OUT	SPL, REG
	LDI	REG, 0b00110111 ; set DDRB value, high for all pins except button
	OUT	DDRB, REG
	LDI	REG, 0b00101000 ; PortB value = all LEDs off, enable pull up for button
	OUT	PORTB, REG
	RCALL	DELAY1 ; wait 10 msec
	LDI 	COUNTER, 0 ; clear counter
	LDI 	REG, (1<<PCIE) ; enable pin change interrupts (general interrupt mask)
	OUT 	GIMSK, REG
	LDI	REG, (1<<BTN_PCINT) ; enable pin change interrupt on BTN_PCINT (pin interrupt mask)
	OUT 	PCMSK, REG
	LDI 	REG, MCUCR_NORMAL ; set MCU control register value
	OUT 	MCUCR, REG
	SEI
MLOOP:	SBRS 	STAT, STAT_SLEEP ; check status, if sleep not enabled, skip sleep-related instructions
	BREQ	MLOOP2
	SEI
	LDI	REG, (1<<BTN_PCINT) ; enable pin change interrupt on BTN_PCINT (pin interrupt mask)
	OUT 	PCMSK, REG
	LDI 	REG, MCUCR_SLEEPE ; set Sleep Enable bit
	OUT 	MCUCR, REG
	SLEEP ; sleep mode
MLOOP2:	RJMP 	MLOOP


; Delay loop (10ms)
DELAY1:	PUSH 	REG
	LDI	REG, 254
DLOOP1:	NOP
	NOP
	DEC	REG
	BRNE 	DLOOP1
	POP 	REG
	RET
